/**
   @author Shin'ichiro Nakaoka
*/

#include "InverseDynamics.h"
#include "Link.h"
#include <cnoid/EigenUtil>

/*
 *  see Kajita et al. Humanoid Robot Ohm-sha,  p.210
 */

using namespace cnoid;

namespace cnoid {

    void calcInverseDynamics(const BodyPtr& body, Link* link, Vector3& out_f, Vector3& out_tau)
    {	
        Link* parent = link->parent;
        if(parent){
            Vector3 sv, sw;

            switch(link->jointType){
            case Link::ROTATIONAL_JOINT:
                sw.noalias() = link->R * link->a;
                sv = link->p.cross(sw);
                break;
            case Link::SLIDE_JOINT:
                sw.setZero();
                sv = link->R * link->d;
                break;
            case Link::FIXED_JOINT:
            default:
                sw.setZero();
                sv.setZero();
                break;
            }
            const Vector3 dsv = parent->w.cross(sv) + parent->vo.cross(sw);
            const Vector3 dsw = parent->w.cross(sw);

            link->dw  = parent->dw  + dsw * link->dq + sw * link->ddq;
            link->dvo = parent->dvo + dsv * link->dq + sv * link->ddq;

            link->sw = sw;
            link->sv = sv;
        }
	
        const Vector3 c = link->R * link->c + link->p;
        Matrix3 I = link->R * link->I * link->R.transpose();
        const Matrix3 c_hat = hat(c);
        I.noalias() += link->m * c_hat * c_hat.transpose();
        const Vector3 P = link->m * (link->vo + link->w.cross(c));
        const Vector3 L = link->m * c.cross(link->vo) + I * link->w;

        out_f   = link->m * (link->dvo + link->dw.cross(c)) + link->w.cross(P);
        out_tau = link->m * c.cross(link->dvo) + I * link->dw + link->vo.cross(P) + link->w.cross(L);

        if(link->child){
            Vector3 f_c;
            Vector3 tau_c;
            calcInverseDynamics(body, link->child, f_c, tau_c);
            out_f   += f_c;
            out_tau += tau_c;
        }

        link->u = link->sv.dot(out_f) + link->sw.dot(out_tau)
            + link->ddq * link->Jm2; // rotor inertia

        if(link->sibling){
            Vector3 f_s;
            Vector3 tau_s;
            calcInverseDynamics(body, link->sibling, f_s, tau_s);
            out_f   += f_s;
            out_tau += tau_s;
        }
    }
}

